#lang racket

(require redex)
(provide (all-defined-out))

(define-language Honu
  [binary-operator + * ^ -]
  [unary-operator ! !!]
  [assoc left right]
  [atom integer]
  [prec integer]
  [combine identity
            [mkbin: binary-operator term]
            [mkun: unary-operator combine]]
  [stack empty [push combine prec stack]]
  [tree-term (literal: integer)
             (id: variable-not-otherwise-mentioned)
             (call: tree-term tree-term ...)
             (list: tree-term ...)
             (block: tree-term ...)
             (arrayref: tree-term tree-term)
             (un: unary-operator tree-term)
             (bin: binary-operator tree-term tree-term)]
  [term atom 
        identifier
        (parens term ...)
        [squares term ...]
        {curlies term ...}
        tree-term]
  [identifier binary-operator
              unary-operator
              variable-not-otherwise-mentioned]

  [sequence (seq term ...)]
  [binding (macro: transformer)
           (binop: prec assoc)
           ;; precedence level and postfix boolean
           (unop: prec fixity)
           (var: identifier)]
  [fixity postfix prefix]
  [transformer a-transformer]

  [boolean #t #f])

;; A combine will either be the identity function or a binary transformer
;; A binary transformer comes with the left hand argument as part
;; of its definition. The right hand argument comes from the argument
;; to do-combine. `do-combine' will create a tree-term with the binary
;; operator, left, and right-hand arguments. Then it will run the old
;; combine that was part of the binary transformer's definition
;; on the result.
(define-metafunction Honu
  do-combine : combine term -> term
  [(do-combine identity term) term]
  [(do-combine (mkbin: binary-operator term_left) term_right)
   (bin: binary-operator term_left term_right)]
  [(do-combine (mkun: unary-operator combine) term)
   (do-combine combine (un: unary-operator term))])

(define-metafunction Honu
  expand : transformer (seq term ...) -> (seq term ...)
  [(expand a-transformer (seq term ...)) (seq term ...)])

;; If the found operator’s prec is greater than the current
;; prec, then a new do-combine function takes a right and
;; applies the operator’s transformer to init and right; the result
;; is then passed to the original do-combine. Meanwhile, prec
;; becomes the found operator’s prec.

;; If the found operator’s prec is less or equal to the current
;; prec, then do-combine is applied to init to produce left; the
;; new do-combine function takes right and applies the operator’s
;; binary transformer to left and right, while prec remains
;; unchanged.

(define-metafunction Honu
  higher : assoc prec prec -> boolean
  [(higher left prec_1 prec_2) #t
   (side-condition (> (term prec_1) (term prec_2)))]
  [(higher right prec_1 prec_2) #t
   (side-condition (>= (term prec_1) (term prec_2)))]
  [(higher assoc prec_1 prec_2) #f])

(define-metafunction Honu
  lower : assoc prec prec -> boolean
  [(lower left prec_1 prec_2) #f
   (side-condition (> (term prec_1) (term prec_2)))]
  [(lower right prec_1 prec_2) #f
   (side-condition (>= (term prec_1) (term prec_2)))]
  [(lower assoc prec_1 prec_2) #t])

(define-metafunction Honu
  lookup : identifier -> binding
  [(lookup +) (binop: 1 left)]
  [(lookup -) (binop: 1 left)]
  [(lookup ^) (binop: 3 right)]
  [(lookup *) (binop: 2 left)]
  [(lookup !) (unop: 5 prefix)]
  [(lookup !!) (unop: 3 postfix)]
  [(lookup identifier) (var: identifier)])

(define-metafunction Honu
  enforest : (seq term ...) combine prec stack ->
             (tuple tree-term (seq term ...))
  [(enforest (seq identifier term_rest ...) combine prec stack_use)
   (enforest (seq term_rest ...)
             (mkun: identifier combine) prec_operator stack_use)
   (where (unop: prec_operator prefix) (lookup identifier))])
