#lang racket

(require redex
         slideshow/pict)

(provide (all-defined-out))

;; Typesetting ----------------------------------------

(define (refit orig new)
  (append (list (between "" (just-before "" (car orig)) (car new)))
          new
          (list (between "" (just-after "" (last new)) (just-after "" (last orig))))))

(define (between s a b)
  (build-lw s
            (+ (lw-line a) (lw-line-span a))
            0
            (+ (lw-column a) (lw-column-span a))
            (max 0 (- (lw-column b)
                      (+ (lw-column a) (lw-column-span a))))))

(define (binary/subscript sep lws)
  (refit 
   lws
   (list
    (list-ref lws 3)
    (just-after (hbl-append (text " " 'roman (default-font-size))
                            (text sep 'roman (default-font-size))
                            (text (symbol->string (lw-e (list-ref lws 2))) `(subscript . roman) (default-font-size))
                            (text " " 'roman (default-font-size)))
                (list-ref lws 3))
    (between "" (list-ref lws 3) (list-ref lws 4))
    (list-ref lws 4))))

(define (taller s)
  (define p (text s (default-style) (default-font-size)))
  (define h (pict-height p))
  (drop-below-ascent (scale (launder p) 1 1.3) (* 0.1 h)))

(define (re-lw new-e lw)
  (build-lw new-e
            (lw-line lw) (lw-line-span lw)
            (lw-column lw) (lw-column-span lw)))

(define (langle) (taller "\u2329"))
(define (rangle) (taller "\u232A"))

(define (atag x) x)

(define (angle-brackets who lws)
  (if (= (length lws) 3)
      (list (re-lw (langle) (car lws))
            (re-lw (atag who) (cadr lws))
            (re-lw (rangle) (caddr lws)))
      (refit
       lws
       (let* ([open (re-lw (langle) (car lws))]
              [prev-tag (cadr lws)]
              [lws (drop (drop-right lws 1) 2)])
         (append (if who
                     (list open
                           (between "" open prev-tag)
                           (re-lw (atag who) prev-tag))
                     (list open (between "" open (car lws))))
                 (if who
                     (comma-ize lws #f)
                     (append
                      (list (car lws)
                            (just-after "," (car lws))
                            (cadr lws))
                      (if (null? (cddr lws))
                          null
                          (list (just-after "," (cadr lws))
                                (caddr lws)))))
                 (list (just-after (rangle) (last lws))))))))

(define (function-call name)
  (lambda (lws)
    (refit lws
           (list* (if name
                      (re-lw name (list-ref lws 1))
                      (list-ref lws 1))
                  (between "(" (list-ref lws 1) (list-ref lws 2))
                  (comma-ize (cddr lws))))))

(define (comma-ize lws [last-is-paren? #t])
  (let loop ([lws lws])
    (cond
     [(or (null? (cdr lws))
          (and last-is-paren? (null? (cddr lws))))
      lws]
     [else
      (list*
       (car lws)
       (between ", " (car lws) (cadr lws))
       (loop (cdr lws)))])))

(define (with-compound-rewriter* thunk l)
  (if (null? l)
      (thunk)
      (with-compound-rewriter
       (car l) (cadr l)
       (with-compound-rewriter* thunk (cddr l)))))

(define (with-atomic-rewriter* thunk l)
  (if (null? l)
      (thunk)
      (with-atomic-rewriter
       (car l) (cadr l)
       (with-atomic-rewriter* thunk (cddr l)))))

(define constructor
  (lambda (lws)
    (angle-brackets (lw-e (cadr lws)) lws)))

(define drop-word
  (lambda (lws) (list* (car lws)
                       (between "" (car lws) (caddr lws))
                       (cddr lws))))

(define (tt str)
  (text str 'modern (floor (* #e0.8 (default-font-size)))))

(define (fun-header arg)
  (hbl-append (tt "function(") arg (tt "){")))

(define (mv str)
  (tt str)
  #;
  (text str `(italic . roman) (default-font-size)))

(define app-arg
  (lambda (l) ((function-call #f) (cons (car l) (cddr l)))))

(define (with-rewrites thunk)
  (with-atomic-rewriter*
   (lambda ()
     (with-compound-rewriter*
      thunk
      (list
       'enforest (function-call 'enforest)
       'lookup (function-call 'lookup)
       'expand app-arg
       'do-combine app-arg
       'seq (lambda (lws) (if (= (length lws) 3)
                              ;; empty sequence:
                              (refit lws (list (re-lw "ε" (car lws))))
                              ;; normal:
                              (refit lws (drop-right (list-tail lws 2) 1))))
       'tuple (lambda (lws) (refit lws (append (list (car lws)
                                                     (between "" (car lws) (caddr lws)))
                                               (comma-ize (drop lws 2)))))
       'push (lambda (lws) 
               (define combine (caddr lws))
               (define prec (cadddr lws))
               (define rest (cadddr (cdr lws)))
               (list (re-lw "[" (car lws))
                     (between "(" (car lws) combine)
                     combine
                     (between ", " combine prec)
                     prec
                     (between ") " prec rest)
                     rest
                     (re-lw "]" (last lws))))
       'parens drop-word
       'squares drop-word
       'curlies drop-word
       'fun: constructor
       'literal: constructor
       'call: constructor
       'bin: constructor
       'un: constructor
       'mkbin: (lambda (lws)
                 (define op (caddr lws))
                 (define left (cadddr lws))
                 (refit lws (list
                             (re-lw (fun-header (mv "t")) (car lws))
                             (constructor
                              (list
                               (re-lw "(" (cadr lws))
                               (between '|bin: | (cadr lws) op)
                               op
                               left
                               (just-after (mv "t") left)
                               (just-after ")" left)))
                             (re-lw (tt "}") (last lws)))))
       'mkun: (lambda (lws)
                (define op (caddr lws))
                (define combine (cadddr lws))
                (refit lws (list
                            (re-lw (fun-header (mv "t")) (car lws))
                            (between (lw-e combine) (car lws) (cadr lws))
                            (between "(" (car lws) (cadr lws))
                            (constructor
                             (list
                              (re-lw "(" (cadr lws))
                              (between '|un: | (cadr lws) op)
                              op
                              (just-after (mv "t") op)
                              (just-after ")" op)))
                            (between ")" op (last lws))
                            (re-lw (tt "}") (last lws)))))
       'unop: constructor
       'binop: constructor
       'macro: constructor
       'list: constructor
       'block: constructor
       'arrayref: constructor
       'higher (lambda (lws) (binary/subscript ">" lws))
       'lower (lambda (lws) (binary/subscript "<" lws)))))
   (list
    'empty (lambda () (text "[]" 'roman (default-font-size))))))

